/*
   Copyright (c) 2004 Chris Eagle (cseagle at nps d0t edu)
   
   Permission is hereby granted, free of charge, to any person obtaining a copy of 
   this software and associated documentation files (the "Software"), to deal in 
   the Software without restriction, including without limitation the rights to 
   use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of 
   the Software, and to permit persons to whom the Software is furnished to do so, 
   subject to the following conditions:
   
   The above copyright notice and this permission notice shall be included in all 
   copies or substantial portions of the Software.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
   FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
 * Disassembler for Nico byte code
 */

#include <stdio.h>
#include <stdlib.h>

enum {
   CPUSH = 0, APUSH = 1, RPUSH = 2, POP = 3, ADD_SP = 4,
   XOR_M = 0x103, ADD_C = 0x104, SUB = 0x105, TEST = 0x106, 
   OR = 0x107, XOR = 0x108, ADD = 0x109, CMP = 0x10A,
   HALT = 0x200, SYSCALL = 0x201, LOAD_C = 0x202, DEC = 0x203, 
   INC = 0x204, CLEAR = 0x205, SCANB = 0x206, TRAP = 0x207, LOAD_D = 0x208, 
   NOP = 0x209, LOAD_B = 0x20A, LOAD_W = 0x20B,
   JMPEQ = 0x300,
   JMP = 0x400, JNZ = 0x401, JZ = 0x402,
   CALL = 0x500, RET = 0x501
};

char instructions[6][12][8] = {
   {"cpush", "apush", "rpush", "pop", "add_sp"},
   {"", "", "", "xor.S", "add", "sub", "test", "or", "xor", "add", "cmp"},
   {"halt", "syscall", "load", "dec", "inc", "clear", "scanb", "trap", 
    "load.d", "nop", "load.b", "load.w"},
   {"jmpeq"},
   {"jmp", "jnz", "jz"},
   {"call", "ret"}
};

int ilen[6][12] = {
   {6, 6, 3, 3, 3},
   {0, 0, 0, 5, 7, 7, 7, 7, 7, 4, 4},
   {2, 6, 7, 3, 3, 3, 5, 2, 4, 3, 4, 4},
   {6},
   {6, 6, 6},
   {6, 2}
};

char sizes[] = "bwd";

typedef struct inst_t {
   unsigned int addr;
   unsigned int op32;
   unsigned char rs, rd, sz, label;
   unsigned short op16, opc;
   char *i;
   struct inst_t *next;
} inst;

inst *lookup(inst *prog, unsigned int addr) {
   inst *i;
   for (i = prog; i; i = i->next) {
      if (i->addr == addr) break;
   }
   return i;
}

void resolve(inst *prog) {
   inst *i, *t;
   for (i = prog; i; i = i->next) {
      switch (i->opc) {
         case JMPEQ: case JMP: case JNZ: case JZ: case CALL:
            t = lookup(prog, i->op32);
            if (t) t->label = (unsigned char) 1;
            //else unknown jump target!!!
            break;
      }
   }
}

int main(int argc, char **argv) {
   FILE *f;
   unsigned int org;
   unsigned short opc;
   unsigned char r, c;
   inst *prog = NULL, *last = NULL, *i;
   
   fprintf(stderr, "What is the program origin: ");
   scanf("%X", &org);
   f = fopen(argv[1], "rb");
   
   while (fread(&opc, sizeof(opc), 1, f) == 1) {
      i = (inst *) calloc(1, sizeof(inst));
      if (last) last->next = i;
      else prog = i;
      last = i;
      c = (unsigned char) (opc >> 8);
      r = (unsigned char) (opc);
      i->opc = (r << 8) + c;
      switch (i->opc) {
         case CPUSH:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 ^= 0x37195411;
            break;
         case APUSH:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x0ADD01337;
            break;
         case RPUSH:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd ^= 0x47;
            break;
         case POP:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd ^= 0x66;
            break;
         case ADD_SP:
            fread(&i->sz, sizeof(i->sz), 1, f);
            i->sz ^= 0x45;
            break;
         case XOR_M:
            fread(&i->rs, sizeof(i->rs), 1, f);
            fread(&i->sz, sizeof(i->sz), 1, f);
            fread(&i->rd, sizeof(i->rd), 1, f);
            break;
         case ADD_C:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 3;
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
         case SUB:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 2;
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
         case TEST:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 5;
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
         case OR:
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 4;
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
         case XOR:
            fread(&i->rd, sizeof(i->rd), 1, f);
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
         case ADD:
            fread(&i->rs, sizeof(i->rs), 1, f);
            i->rs -= 1;
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 3;
            break;
         case CMP:
            fread(&i->rs, sizeof(i->rs), 1, f);
            i->rs -= 2;
            fread(&i->rd, sizeof(i->rd), 1, f);
            i->rd -= 1;
            break;
         case SYSCALL:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x0FEA731DE;
            break;
         case LOAD_C:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x0AEFDED04;
            fread(&i->rd, sizeof(i->rd), 1, f);
            break;
         case DEC: case INC: case CLEAR: case NOP:
            fread(&i->rd, sizeof(i->rd), 1, f);
            break;
         case SCANB:
            fread(&i->sz, sizeof(i->sz), 1, f);
            fread(&i->op16, sizeof(i->op16), 1, f);
            break;
         case LOAD_D: case LOAD_B: case LOAD_W:
            fread(&i->rs, sizeof(i->rs), 1, f);
            fread(&i->rd, sizeof(i->rd), 1, f);
            break;
         case JMPEQ:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 -= 0x31337;
            break;
         case JMP:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x0DEADEAD;
            break;
         case JNZ:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x00E1BA3E;
            break;
         case JZ:
            fread(&i->op32, sizeof(i->op32), 1, f);
            i->op32 += 0x0E1BA41;
            break;
         case CALL:
            fread(&i->op32, sizeof(i->op32), 1, f);
            break;
      }
      last->i = instructions[r][c];
      last->addr = org;
      org += ilen[r][c];
   }
   fclose(f);
   resolve(prog);
   if (prog) prog->label = 1;
   while ((i = prog) != NULL) {
      if (i->label) printf("loc_%8.8X:\n", i->addr);
      printf("      ");
      switch (i->opc) {
         case CPUSH: case APUSH:
            printf("%-6s 0x%X\n", i->i, i->op32);
            break;
         case RPUSH: case POP: case DEC: case INC: case CLEAR: case NOP:
            printf("%-6s R%d\n", i->i, i->rd);
            break;
         case ADD_SP:
            printf("%-6s 0x%X\n", i->i, i->sz);
            break;
         case XOR_M:
            i->i[4] = sizes[i->sz];
            printf("%-6s [R%d],R%d\n", i->i, i->rd, i->rs);
            break;
         case ADD_C: case SUB: case TEST: case OR: case XOR: case LOAD_C:
            printf("%-6s R%d,0x%X\n", i->i, i->rd, i->op32);
            break;
         case ADD: case CMP:
            printf("%-6s R%d,R%d\n", i->i, i->rd, i->rs);
            break;
         case SCANB:
            printf("%-6s 0x%2.2X,0x%X\n", i->i, i->sz, i->op16);
            break;
         case HALT: case TRAP: case RET:
            printf("%s\n", i->i);
            break;
         case LOAD_D: case LOAD_B: case LOAD_W:
            printf("%-6s R%d,[R%d]\n", i->i, i->rd, i->rs);
            break;
         case SYSCALL: case JMPEQ: case JMP: case JNZ: case JZ: case CALL:
            printf("%-6s loc_%8.8X\n", i->i, i->op32);
            break;
      }
      prog = prog->next;
      free(i);
   }
   return 0;
}

